import torch.nn as nn
import torch.nn.functional as F
import einops


class ShuffleLayer(nn.Module):
    def __init__(self, hidden_dim, shuffle_dim):
        super().__init__()
        self.q = nn.Linear(hidden_dim, shuffle_dim)
        self.k = nn.Linear(hidden_dim, shuffle_dim)
        self.v = nn.Linear(hidden_dim, shuffle_dim)


    def forward(self, x, gamma, *args, **kwargs):
        q = self.q(x)
        k = self.k(x)
        v = x

        attention = F.softmax(q @ k.transpose(1, 2), dim=-2)
        # ik WIL sharpness, dus niet delen door (dim)**.5

        return gamma * (attention @ v) + (1 - gamma) * x


class BucketAttentionLayer(nn.Module):
    def __init__(self, hidden_dim, heads, n_buckets):
        super().__init__()
        self.q = nn.Linear(hidden_dim, hidden_dim)
        self.k = nn.Linear(hidden_dim, hidden_dim)
        self.v = nn.Linear(hidden_dim, hidden_dim)
        self.heads = heads
        self.n_buckets = n_buckets


    def forward(self, x, *args, **kwargs):
        h = self.heads

        q = self.q(x)
        k = self.k(x)
        v = self.v(x)

        q = einops.rearrange(q, 'b (n_buckets bucket_size) (h h_dim) -> b n_buckets h bucket_size h_dim',
                             n_buckets=self.n_buckets, h=h)
        k = einops.rearrange(k, 'b (n_buckets bucket_size) (h h_dim) -> b n_buckets h bucket_size h_dim',
                             n_buckets=self.n_buckets, h=h)
        v = einops.rearrange(v, 'b (n_buckets bucket_size) (h h_dim) -> b n_buckets h bucket_size h_dim',
                             n_buckets=self.n_buckets, h=h)

        attention = F.softmax(q @ k.transpose(3, 4), dim=-1)
        attended = attention @ v

        attended = einops.rearrange(attended, 'b n_buckets h bucket_size h_dim -> b (n_buckets bucket_size) (h h_dim)')
        return attended + x


class ShuffleFormerLayer(nn.Module):
    def __init__(self, seq_len, hidden_dim, shuffle_dim, heads, n_buckets):
        super().__init__()
        assert seq_len % n_buckets == 0
        self.shuffle = ShuffleLayer(hidden_dim, shuffle_dim)
        self.bucket_attention = BucketAttentionLayer(hidden_dim, heads, n_buckets)


    def forward(self, x, gamma, *args, **kwargs):
        x = self.shuffle(x, gamma)
        x = self.bucket_attention(x)
        return x


class FeedforwardLayer(nn.Module):
    def __init__(self, hidden_dim, ff_dim=None, act=nn.GELU()):
        super().__init__()
        if not ff_dim:
            ff_dim = 4 * hidden_dim
        self.up_scale = nn.Linear(hidden_dim, ff_dim)
        self.act = act
        self.down_scale = nn.Linear(ff_dim, hidden_dim)

    def forward(self, x, *args, **kwargs):
        return x + self.down_scale(self.act(self.up_scale(x)))


class PreNorm(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.norm = nn.LayerNorm(dim)

    def forward(self, x, *args, **kwargs):
        return self.norm(x)


class Attention(nn.Module):
    def __init__(self, hidden_dim, heads):
        super().__init__()
        self.q = nn.Linear(hidden_dim, hidden_dim)
        self.k = nn.Linear(hidden_dim, hidden_dim)
        self.v = nn.Linear(hidden_dim, hidden_dim)
        self.heads = heads

    def forward(self, x, y=None, *args, **kwargs):
        if y is None:
            y = x
        h = self.heads

        q = self.q(x)
        q = einops.rearrange(q, 'b n1 (h h_dim) -> b h n1 h_dim', h=h)
        k = self.k(y)
        k = einops.rearrange(k, 'b n2 (h h_dim) -> b h n2 h_dim', h=h)
        v = self.v(y)
        v = einops.rearrange(v, 'b n2 (h h_dim) -> b h n2 h_dim', h=h)

        attention = F.softmax(q @ k.transpose(2, 3), dim=-1)

        attended = einops.rearrange(attention @ v, 'b h n h_dim -> b n (h h_dim)')

        return x + attended


















import torch
import tiggy.rl_utils as utils
import itertools
import numpy as np
import torch.nn.functional as F


class Trainer:
    def __init__(self, env, agent, parallel_size, experience_threshold, batch_size=256, decay_factor=.97):
        self.env = env
        self.agent = agent
        self.parallel_size = parallel_size
        self.envs = [self.env() for _ in range(parallel_size)]
        self.repr_storage = [[] for env in self.envs]
        self.move_storage = [env.reset() for env in self.envs]
        self.big_storage = []
        self.experience_threshold = experience_threshold
        self.batch_size = batch_size
        self.decay_factor = decay_factor


    def step(self):
        inputs = utils.unpack(self.move_storage)
        with torch.no_grad():
            moves = self.agent.move(torch.from_numpy(inputs['obs']).float(), torch.from_numpy(inputs['legal']))
            values = self.agent.evaluate(torch.from_numpy(inputs['obs']).float()).cpu().numpy()

        self.move_storage = [env.move(move) for env, move in zip(self.envs, moves)]
        unpacked_env_outputs = utils.unpack(self.move_storage)
        for i, (obs, legal, reward, is_final, move, value) in enumerate(zip(inputs['obs'], inputs['legal'], unpacked_env_outputs['reward'], unpacked_env_outputs['is_final'], moves, values)):
            self.repr_storage[i] += [{'obs':obs, 'legal':legal, 'reward':reward, 'is_final':is_final, 'move':move, 'value':value}]
            if is_final:
                self.big_storage += [self.repr_storage[i]]
                self.repr_storage[i] = []
                if len(self.big_storage) >= self.experience_threshold:
                    self.train()

    def train(self):
        big_storage = utils.decay_rewards(self.big_storage, self.decay_factor)
        big_storage = [record for storage in big_storage for record in storage]
        unpacked_big_storage = utils.unpack(big_storage)

        x = torch.from_numpy(unpacked_big_storage['obs']).float().cuda()
        rewards = torch.from_numpy(np.array(unpacked_big_storage['reward'])).cuda()
        moves = torch.from_numpy(np.array(unpacked_big_storage['move'])).long().cuda()
        legal = torch.from_numpy(unpacked_big_storage['legal']).cuda()

        with torch.no_grad():
            advantages = rewards - torch.reshape(self.agent.evaluate(x), (-1,))
            std, mean = torch.std_mean(advantages)
            advantages = (advantages-mean)/std

        # with torch.enable_grad():
        p = self.agent.use_policy(x, legal)
        loss = torch.mean(F.cross_entropy(p, moves, reduction='none') * advantages)
        self.agent.policy_optimizer.zero_grad()
        loss.backward()
        self.agent.policy_optimizer.step()

        v = self.agent.evaluate(x)
        loss = F.mse_loss(v, torch.reshape(rewards, (-1, 1)).float())
        self.agent.value_optimizer.zero_grad()
        loss.backward()
        self.agent.value_optimizer.step()

        self.big_storage = []



from .transformer_layers import *
import torch.nn as nn
import torch
import einops


class ShuffleFormer(nn.Module):
    def __init__(self, n_vocab, depth, hidden_dim, seq_len, shuffle_dim, heads, n_buckets):
        super().__init__()
        self.pos_embedding = nn.Parameter(torch.zeros((seq_len, hidden_dim)))
        self.embedding = nn.Embedding(n_vocab, hidden_dim)
        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers += [PreNorm(hidden_dim), ShuffleFormerLayer(seq_len, hidden_dim, shuffle_dim, heads, n_buckets),
                            PreNorm(hidden_dim), FeedforwardLayer(hidden_dim)]

    def forward(self, x, gamma):
        b, *_ = x.shape
        x = self.embedding(x)
        x += einops.repeat(self.pos_embedding, 'n d -> b n d', b=b)
        for layer in self.layers:
            x = layer(x, gamma)

        return x


class Transformer(nn.Module):
    def __init__(self, depth, hidden_dim, seq_len, heads, n_vocab=None, input_dim=None, use_cls_token=False, out_dim=None):
        super().__init__()
        self.pos_embedding = nn.Parameter(torch.zeros((seq_len, hidden_dim)))
        if n_vocab is not None:
            self.embedding = nn.Embedding(n_vocab, hidden_dim)
        elif input_dim is not None:
            self.embedding = nn.Linear(input_dim, hidden_dim)
        self.use_cls_token = use_cls_token
        if use_cls_token:
            self.cls_token = nn.Parameter(torch.zeros((hidden_dim)))

        self.out_projection = nn.Linear(hidden_dim, out_dim) if out_dim is not None else nn.Identity()

        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers += [PreNorm(hidden_dim), Attention(hidden_dim, heads),
                            PreNorm(hidden_dim), FeedforwardLayer(hidden_dim)]

    def forward(self, x):
        b, *_ = x.shape
        x = self.embedding(x)
        x += einops.repeat(self.pos_embedding, 'n d -> b n d', b=b)
        if self.use_cls_token:
            x = torch.cat([x, einops.repeat(self.cls_token, 'b -> b n d', b=b, n=1)], dim=1)
        for layer in self.layers:
            x = layer(x)

        if self.use_cls_token:
            return self.out_projection(x[:, -1])
        return self.out_projection(x[:, -1])


import tiggy.rl_utils as utils
import numpy as np
import torch.nn as nn
import torch
import torch.nn.functional as F

class Agent(nn.Module):
    def __init__(self, policy_net, policy_net_settings, value_net, value_net_settings):
        super().__init__()
        self.policy_net = policy_net(**policy_net_settings).cuda()
        self.value_net = value_net(**value_net_settings).cuda()
        self.policy_optimizer = torch.optim.AdamW(self.policy_net.parameters(), 3e-4)
        self.value_optimizer = torch.optim.AdamW(self.value_net.parameters(), 3e-4)

    def move(self, x, legal):
        p = self.use_policy(x, legal).cpu().numpy()
        moves = utils.get_moves(p).astype(np.int64)
        return moves

    def use_policy(self, x, legal):
        logits = self.policy_net(x.cuda())
        logits[~legal.cuda()] = float('-inf')
        p = F.softmax(logits, dim=-1)
        return p

    def evaluate(self, x):
        value = self.value_net(x.cuda())
        return value


import numpy as np
import numba as nb
from typing import Tuple


class TicTacToe:
    def __init__(self):
        self.board = np.zeros((3, 3))
        self.turn = 1

    def get_legal(self):
        return self.board.flatten() == 0

    def move(self, move):
        self.board[move//3, move%3] = self.turn
        final, has_won = is_final(self.board, self.turn)
        self.turn = 3 - self.turn
        obs = self.get_repr()
        if final:
            self.reset()
        return {'obs':obs, 'legal':self.get_legal(), 'is_final':final, 'reward':(reward := int(has_won))}

    def get_repr(self):
        return get_input(self.board, self.turn)

    def reset(self):
        self.board = np.zeros((3, 3))
        self.turn = 1
        return {'obs':self.get_repr(), 'legal':self.get_legal()}

@nb.njit
def is_final(board, turn) -> Tuple[bool, bool]:
    for i in range(3):
        if np.all(board[i, :] == turn) or np.all(board[:, i] == turn):
            return True, True
    if board[0, 0] == turn and board[1, 1] == turn and board[2, 2] == turn:
        return True, True
    if board[2, 0] == turn and board[1, 1] == turn and board[0, 2] == turn:
        return True, True
    if 0 not in board:
        return True, False
    return False, False

@nb.njit
def get_input(board, turn):
    in_arr = np.zeros((3, 3, 2))
    for i in range(3):
        for j in range(3):
            in_arr[i, j] = [board[i, j] == turn, board[i, j] == 3 - turn]
    return in_arr.reshape((-1, 2))


import numpy as np
import numba as nb
from typing import List, Dict
from collections.abc import Iterable
import torch
import random


# @nb.njit
def get_moves(p):
    moves = np.zeros(len(p))
    for i, moveset in enumerate(p):
        chance = random.random()
        for j in range(len(moveset)):
            chance -= moveset[j]
            if chance <= 0:
                moves[i] = j
                break
    return moves

type_dict = {np.ndarray: np.array}


def unpack(env_outputs: List[Dict]) -> Dict[str, List]:
    keywords = env_outputs[0].keys()
    rewrite = {keyword:[] for keyword in keywords}
    for env_output in env_outputs:
        for key, value in env_output.items():
            rewrite[key] += [value]

    for key, value in rewrite.items():
        if isinstance(value[0], Iterable):
            rewrite[key] = type_dict[type(value[0])](value)
    return rewrite


def decay_rewards(env_outputs: List[List[Dict]], decay_factor) -> List[Dict]:
    for env_output in env_outputs:
        unpacked_env_output = unpack(env_output)
        rewards = unpacked_env_output['reward']
        decayed_rewards = np.array(rewards, dtype=np.float64)
        for i in reversed(range(len(rewards) - 1)):
            decayed_rewards[i] = rewards[i] + decay_factor * decayed_rewards[i+1]

        for i, moment in enumerate(env_output):
            moment['reward'] = decayed_rewards[i]

    return env_outputs



from setuptools import setup, find_packages

setup(
  name = 'tiggy',
  packages = find_packages(),
  version = '0.5',
  license='GPL v3.0',
  description = 'tiggy',
  install_requires=[
    'numpy',
    'einops',
    'torch',
    'numba',
  ],
)